/* SPDX-License-Identifier: BSD-2-Clause */

/* Dummy Capsicum helper implementation */

#ifndef CAPSICUM_HELPERS_H
#define CAPSICUM_HELPERS_H

#define CAPH_READ		0x0001
#define CAPH_WRITE		0x0002
#define CAPH_IGNORE_EBADF	0x0004

static int caph_limit_stream(int fd, int flags)
{
    (void)fd;
    (void)flags;
    return 0;
}

static int caph_limit_stdio(void) { return 0; }

static int caph_enter(void) { return 0; }

#endif /* CAPSICUM_HELPERS_H */
